#include <stdint.h>

#ifndef _CLASSFILE_H
#define _CLASSFILE_H 1

#define CLASSPATH "CLASSPATH"
extern const uint32_t magic;

/* Types of constant_pool */
typedef enum ECONSTANT
{
    Class = 7,
    Fieldref = 9,
    Methodref = 10,
    InterfaceMethodref = 11,
    String = 8,
    Integer = 3,
    Float = 4,
    Long = 5,
    Double = 6,
    NameAndType = 12,
    Utf8 = 1
} TCONSTANT;

typedef struct SCONSTANT_Class_info
{
    /* uint8_t tag; CONSTANT_Class = 7 */
    uint16_t name_index;
} TCONSTANT_Class_info;

typedef struct SCONSTANT_Fieldref_info
{
    /* uint8_t tag; CONSTANT_Fieldref = 9 */
    uint16_t class_index;
    uint16_t name_and_type_index;
} TCONSTANT_Fieldref_info;

typedef struct SCONSTANT_Methodref_info
{
    /* uint8_t tag; CONSTANT_Methodref = 10 */
    uint16_t class_index;
    uint16_t name_and_type_index;
} TCONSTANT_Methodref_info;

typedef struct SCONSTANT_InterfaceMethodref_info
{
    /* uint8_t tag; CONSTANT_InterfaceMethodref = 11 */
    uint16_t class_index;
    uint16_t name_and_type_index;
} TCONSTANT_InterfaceMethodref_info;

typedef struct SCONSTANT_String_info
{
    /* uint8_t tag; CONSTANT_String = 8 */
    uint16_t string_index;
} TCONSTANT_String_info;

typedef struct SCONSTANT_Integer_info
{
    /* uint8_t tag; CONSTANT_Integer = 3 */
    uint32_t bytes;
} TCONSTANT_Integer_info;

typedef struct SCONSTANT_Float_info
{
    /* uint8_t tag; CONSTANT_Float = 4 */
    uint32_t bytes;
} TCONSTANT_Float_info;

typedef struct SCONSTANT_Long_info
{
    /* uint8_t tag; CONSTANT_Long = 5 */
    uint32_t /*high_*/bytes;
    /* uint32_t low_bytes; */
} TCONSTANT_Long_info;

typedef struct SCONSTANT_Double_info
{
    /* uint8_t tag; CONSTANT_Double = 6 */
    uint32_t /*high_*/bytes;
    /* uint32_t low_bytes; */
} TCONSTANT_Double_info;

typedef struct SCONSTANT_NameAndType_info
{
    /* uint8_t tag; CONSTANT_NameAndType = 12 */
    uint16_t name_index;
    uint16_t descriptor_index;
} TCONSTANT_NameAndType_info;

typedef struct SCONSTANT_Utf_info
{
    /* uint8_t tag; CONSTANT_Utf8 = 1 */
    uint16_t length;
    uint8_t * bytes;
} TCONSTANT_Utf8_info;

typedef struct Scp_info
{
    /*uint8_t*/TCONSTANT tag;
    union {
    	TCONSTANT_Class_info class_info;
    	TCONSTANT_Fieldref_info fieldref_info;
    	TCONSTANT_Methodref_info methodref_info;
    	TCONSTANT_InterfaceMethodref_info interfacemethodref_info;
    	TCONSTANT_String_info string_info;
    	TCONSTANT_Integer_info integer_info;
    	TCONSTANT_Float_info float_info;
    	TCONSTANT_Long_info long_info;
    	TCONSTANT_Double_info double_info;
    	TCONSTANT_NameAndType_info nameandtype_info;
    	TCONSTANT_Utf8_info utf8_info;
    } info;
} Tcp_info;

/* Access flags for fields and methods */
typedef enum EACC
{
    PUBLIC = 0x0001,
    PRIVATE = 0x0002,
    PROTECTED = 0x0004,
    STATIC = 0x0008,
    FINAL = 0x0010,
    SYNCHRONIZED = 0x0020,
    VOLATILE = 0x0040,
    TRANSIENT = 0x0080,
    NATIVE = 0x0100,
    ABSTARCT = 0x0400,
    STRICT = 0x0800
} TACC;

/* Attributes */
#define ISourceFile_attribute "SourceFile"
typedef struct SSourceFile_attribute
{
    /* uint16_t attribute_name_index; SourceFile */
    uint16_t sourcefile_index;
} TSourceFile_attribute;

#define IConstantValue_attribute "ConstantValue"
typedef struct SConstantValue_attribute
{
    /* uint16_t attribute_name_index; ConstantValue */
    uint16_t constantvalue_index;
} TConstantValue_attribute;

typedef struct SJexception
{
    uint16_t start_pc;
    uint16_t end_pc;
    uint16_t handler_pc;
    uint16_t catch_type;
} TJexceptions;

struct Sattribute_info;

#define ICode_attribute "Code"
typedef struct SCode_attribute
{
    /* uint16_t attribute_name_index; Code */
    uint16_t max_stack;
    uint16_t max_locals;
    uint32_t code_length;
    uint8_t *code;
    uint16_t exception_table_length;
    TJexceptions *exception_table;
    uint16_t attributes_count;
    struct Sattribute_info *attributes;
} TCode_attribute;

#define IExceptions_attribute "Exceptions"
typedef struct SExceptions_attribute
{
    /* uint16_t attribute_name_index; Exceptions */
    uint16_t number_of_exceptions;
    uint16_t *exception_index_table;
} TExceptions_attribute;

typedef struct SJlinenumbertable
{
    uint16_t start_pc;
    uint16_t line_number;
} TJlinenumbertable;

#define ILineNumberTable_attribute "LineNumberTable"
typedef struct SLineNumberTable_attribute
{
    /* uint16_t attribute_name_index; LineNumberTable */
    uint16_t line_number_table_length;
    TJlinenumbertable *line_number_table;
} TLineNumberTable_attribute;

typedef struct SJlocalvariabletable
{
    uint16_t start_pc;
    uint16_t length;
    uint16_t name_index;
    uint16_t descriptor_index;
    uint16_t index;
} TJlocalvariabletable;

#define ILocalVariableTable_attribute "LocalVariableTable"
typedef struct SLocalVariableTable_attribute
{
    /* uint16_t attribute_name_index; LocalVariableTable */
    uint16_t local_variable_table_length;
    TJlocalvariabletable *local_variable_table;
} TLocalVariableTable_attribute;

#define IInnerClasses_attribute "InnerClasses"
typedef struct Sinner_class_info
{
    uint16_t inner_class_info_index;
    uint16_t outer_class_info_index;
    uint16_t inner_name_index;
    uint16_t inner_class_access_flags;
} Tinner_class_info;

typedef struct SInnerClasses_attribute
{
    /* uint16_t attribute_name_index; InnerClasses */
    uint16_t number_of_classes;
    struct Sinner_class_info *classes;
} TInnerClasses_attribute;

#define ISynthetic_attribute "Synthetic"
typedef struct SSynthetic_attribute
{
    /* uint16_t attribute_name_index; Synthetic */
    uint16_t dummy;
} TSynthetic_attribute;

#define IDeprecated_attribute "Deprecated"
typedef struct SDeprecated_attribute
{
    /* uint16_t attribute_name_index; Deprecated */
    uint16_t dummy;
} TDeprecated_attribute;

typedef struct Sattribute_info
{
    uint16_t attribute_name_index;
    uint32_t attribute_length;
    union {
    	TSourceFile_attribute sourcefile;
    	TConstantValue_attribute constantvalue;
    	TCode_attribute code;
    	TExceptions_attribute exceptions;
    	TLineNumberTable_attribute linenumbertable;
    	TLocalVariableTable_attribute localvariabletable;
        TInnerClasses_attribute innerclasses;
        TSynthetic_attribute synthetic;
        TDeprecated_attribute deprecated;
    } info;
} Tattribute_info;

/* Field */
typedef struct Sfield_info
{
    /*uint16_t*/TACC access_flags;
    uint16_t name_index;
    uint16_t descriptor_index;
    uint16_t attributes_count;
    Tattribute_info *attributes;
} Tfield_info;

/* Methods */
typedef struct Smethod_info
{
    /*uint16_t*/TACC access_flags;
    uint16_t name_index;
    uint16_t descriptor_index;
    uint16_t attributes_count;
    Tattribute_info *attributes;
} Tmethod_info;

/* Class header */
typedef struct SClassFile
{
    uint32_t magic;
    uint16_t minor_version;
    uint16_t major_version;
    uint16_t constant_pool_count;
    Tcp_info  *constant_pool;
    uint16_t access_flags;
    uint16_t this_class;
    uint16_t super_class;
    uint16_t interfaces_count;
    uint16_t *interfaces;
    uint16_t fields_count;
    Tfield_info *fields;
    uint16_t methods_count;
    Tmethod_info *methods;
    uint16_t attributes_count;
    Tattribute_info *attributes;
} TClassFile;

extern TClassFile classFile;

void readCpInfo(int fd, Tcp_info *info, int *i);
void readAttributeInfo(int fd, Tattribute_info *info);
void readFieldsInfo(int fd, Tfield_info *info);
void readMethodInfo(int fd, Tmethod_info *info);
void readClassFile(int fd);

#endif
/* vim: autoindent expandtab tabstop=4 shiftwidth=4
*/
